/* tfix2: fixing ID values and eliminating improper macro_call tags 
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: tfix2.c,v 1.8 2006/11/01 08:19:04 gondow Exp $ */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>
#include <libxml/xmlreader.h>
#include <libxml/xmlwriter.h>

#include "misc.h"
#include "list.h"
/* ==================================================================== */
struct fentry {
    char   *file_id;
    struct fentry *next;
};

static xmlTextReaderPtr reader;
static xmlTextWriterPtr writer;

enum { HASH_SIZE = 1024, };
static struct fentry hash_table [HASH_SIZE];

enum { MC_NONE = 0, MC_OBJ = 1, MC_FUNC = 2, MC_PARAM = 3, MC_DEFINED = 4, };
static struct CDLL mc_stack;

/* ==================================================================== */
static void fentry_dump (void);
static int  hash (char *id);
static void fentry_insert (char *id, int i);
static int  check_id (char *id);
static int  mc_type_to_int (char *type);
static void output_attr (void);
static void output_is_defined_tag (char *mc_tag, char *mc_name, int mc_type);
static void output_other_tag (char *name);
static void store_macro_call_tag (char *name, char **mc_tag, char **mc_name,
				  int *mc_type);
static void tfix (const char *outfile);
/* ==================================================================== */
static void
fentry_dump  (void)
{
    int i;
    struct fentry *p;
    for (i = 0; i < HASH_SIZE; i++) {
	fprintf (stderr, "%d: ", i);
	for (p = hash_table [i].next; p != NULL; p = p->next) {
	    fprintf (stderr, "id=%s, ", p->file_id);
	}
	fprintf (stderr, "\n");
    }
}

static int
hash (char *id)
{
    int h = 0;
    while (*id) {
	h = (h * 37 + *id++) % HASH_SIZE;
    }
    return h;
}

static void
fentry_insert (char *id, int i)
{
    struct fentry *p = malloc (sizeof (struct fentry));
    p->file_id = strdup (id);
    p->next = hash_table [i].next;
    hash_table [i].next = p;
}

static int
check_id (char *id)
{
    struct fentry *p;
    int i = hash (id);
    for (p = hash_table [i].next; p != NULL; p = p->next)
	if (!strcmp (id, p->file_id))
	    return 0;
    fentry_insert (id, i);
    return 1;
}

static int
mc_type_to_int (char *type)
{
    if (!strcmp (type, "obj"))          return MC_OBJ;
    else if (!strcmp (type, "func"))    return MC_FUNC;
    else if (!strcmp (type, "param"))   return MC_PARAM;
    else if (!strcmp (type, "defined")) return MC_DEFINED;
    else                                return MC_NONE;
}

static void
output_attr (void)
{
    if (xmlTextReaderHasAttributes (reader)) {
	while (xmlTextReaderMoveToNextAttribute (reader)) {
	    char *name = xmlTextReaderName (reader);
	    char *val  = xmlTextReaderValue (reader);
	    if (strcmp (name, "id") || check_id (val)) {
		xmlTextWriterWriteAttribute (writer, name, val);
	    }
	    xmlFree (name);
	    xmlFree (val);
	}
	xmlTextReaderMoveToElement (reader);
    }
}

static void
output_is_defined_tag (char *mc_tag, char *mc_name, int mc_type)
{
    char *is_defined = xmlTextReaderReadString (reader);
    char *is_defined2 = estrdup (is_defined);
    int depth = xmlTextReaderDepth (reader);
    sscanf (is_defined, " %s", is_defined2);

    assert (mc_tag != NULL);
    assert (mc_name != NULL);

    if (!strcmp (is_defined2, "1")) { /* real macro */
	xmlTextReaderPtr mem_reader;
	char *name, *val;

	mem_reader = xmlReaderForMemory (mc_tag, strlen (mc_tag),
					 NULL, NULL, 0);
#if 1
	xmlTextReaderSetParserProp (reader, XML_PARSER_SUBST_ENTITIES, 0);
	xmlTextReaderSetParserProp (reader, XML_PARSER_VALIDATE, 0);
	xmlTextReaderSetParserProp (reader, XML_PARSER_LOADDTD, 0);
#endif
	xmlTextReaderRead (mem_reader);
	name = xmlTextReaderName  (mem_reader);
	val  = xmlTextReaderValue (mem_reader);

	xmlTextWriterStartElement (writer, name);

	if (xmlTextReaderHasAttributes (mem_reader)) {
	    while (xmlTextReaderMoveToNextAttribute (mem_reader)) {
		char *name = xmlTextReaderName  (mem_reader);
		char *val  = xmlTextReaderValue (mem_reader);
		/* we assume <macro_call> does NOT have 'id' attribute */
		xmlTextWriterWriteAttribute (writer, name, val);
		xmlFree (name); xmlFree (val);
	    }
	}

	output_other_tag ("is_defined");
	xmlFree (name); xmlFree (val);
    } else { /* not macro */
	CDLL_push (&mc_stack, depth - 1, mc_type);
	xmlTextWriterWriteFormatRaw (writer, "%s", mc_name);
	if (mc_type == MC_FUNC)
	    xmlTextWriterWriteFormatRaw (writer, "(");
    }
    
    xmlFree (is_defined);
    EFREE (is_defined2); EFREE (mc_tag); EFREE (mc_name);
}

static void
output_other_tag (char *name)
{
    xmlTextWriterStartElement (writer, name);
    output_attr ();
    if (xmlTextReaderIsEmptyElement (reader))
	xmlTextWriterEndElement (writer);
}

static void
store_macro_call_tag (char *name, char **mc_tag, char **mc_name, int *mc_type)
{
    xmlTextWriterPtr mem_writer;
    xmlBufferPtr xbuf = xmlBufferCreate ();

    assert (xbuf != NULL);
    mem_writer = xmlNewTextWriterMemory (xbuf, 0);
    assert (mem_writer != NULL);

    xmlTextWriterStartElement (mem_writer, name);

    if (xmlTextReaderHasAttributes (reader)) {
	while (xmlTextReaderMoveToNextAttribute (reader)) {
	    char *name = xmlTextReaderName (reader);
	    char *val  = xmlTextReaderValue (reader);
	    if (!strcmp (name, "type"))
		*mc_type = mc_type_to_int (val);
	    if (!strcmp (name, "name"))
		*mc_name = estrdup (val);
	    if (strcmp (name, "id") || check_id (val)) {
		xmlTextWriterWriteAttribute (mem_writer, name, val);
	    } 
	    xmlFree (name); xmlFree (val);
	}
	xmlTextReaderMoveToElement (reader);
    }

    xmlTextWriterEndElement (mem_writer);
    xmlTextWriterEndDocument(mem_writer);
    xmlFreeTextWriter (mem_writer);
    *mc_tag =  estrdup (xmlBufferContent (xbuf));
    xmlBufferFree (xbuf);
}

static void
tfix (const char *outfile)
{
    int ret;
    int is_first_arg = 0;

    reader = xmlReaderForFd (STDIN_FILENO, NULL, NULL, 0);

    if (reader == NULL) {
	fprintf (stderr, "cannot open: stdin\n");
	exit (1);
    }

    ret = xmlTextReaderSetParserProp (reader, XML_PARSER_SUBST_ENTITIES, 0);
    assert (ret == 0);
    ret = xmlTextReaderSetParserProp (reader, XML_PARSER_VALIDATE, 0);
    assert (ret == 0);
    ret = xmlTextReaderSetParserProp (reader, XML_PARSER_LOADDTD, 0);
    assert (ret == 0);
#if 0
    ret = xmlTextReaderSetParserProp (reader, XML_PARSER_DEFAULTATTRS, 0);
    assert (ret == 0);
#endif

    /* ???: how can I open a writer via STDOUT_FILENO */
    writer = xmlNewTextWriterFilename (outfile, 0);
    if (writer == NULL) {
	fprintf (stderr, "cannot open: %s\n", outfile);
	exit (1);
    }

    xmlTextWriterWriteString (writer, "<?xml version=\"1.0\"?>\n");

    while(xmlTextReaderRead (reader)) {
    again:;
	char *name = xmlTextReaderName (reader);
	char *val  = xmlTextReaderValue (reader);
	int  depth = xmlTextReaderDepth (reader);
	int  type;

	char *mc_tag;
	char *mc_name;
	int  mc_type;

	switch (type = xmlTextReaderNodeType (reader)) {
	case XML_READER_TYPE_ELEMENT:
	    if (!strcmp (name, "macro_call")) {
		store_macro_call_tag (name, &mc_tag, &mc_name, &mc_type);
#if 0
		fprintf (stderr, "macro_call_tag=|%s|\n", mc_tag);
#endif
	    } else if (!strcmp (name, "is_defined")) {
		output_is_defined_tag (mc_tag, mc_name, mc_type);
	    } else {
		int mc_depth, mc_type;
		CDLL_top (&mc_stack, &mc_depth, &mc_type);
		if (depth != mc_depth + 1 && depth != mc_depth + 2) {
		    output_other_tag (name);
		} else {
		    if (!strcmp (name, "macro_body")
			|| !strcmp (name, "macro_call")) {
			xmlTextReaderNext (reader); /* skip the subtree */
			goto again;
		    }
		    if (!strcmp (name, "double")) {
			is_first_arg = 1;
		    }
		    if (!strcmp (name, "arg")) {
			if (!is_first_arg)
			    xmlTextWriterWriteFormatRaw (writer, ", ");
			is_first_arg = 0;
		    }
		    assert (!strcmp (name, "arg")
			    || !strcmp (name, "bare")
			    || !strcmp (name, "double"));
		}
	    }

	    break;

	case XML_READER_TYPE_END_ELEMENT:
	    if (!strcmp (name, "macro_call")) {
		int mc_depth, mc_type;

		CDLL_top (&mc_stack, &mc_depth, &mc_type);
		if (depth == mc_depth) {
		    struct CDLL *cdll = CDLL_pop (&mc_stack);
		    if (mc_type == MC_FUNC)
			xmlTextWriterWriteFormatRaw (writer, ")");
		    EFREE (cdll);
		} else {
		    xmlTextWriterEndElement (writer);
		}
	    } else {
		int mc_depth, mc_type;

		CDLL_top (&mc_stack, &mc_depth, &mc_type);
		if (depth != mc_depth + 1 && depth != mc_depth + 2) {
		    xmlTextWriterEndElement (writer);
		} else {
		    if (!strcmp (name, "arg")) {
			is_first_arg = 0;
		    }
		    assert (!strcmp (name, "arg")
			    || !strcmp (name, "is_defined")
			    || !strcmp (name, "bare")
			    || !strcmp (name, "double"));
		}
	    }
	    break;

	case XML_READER_TYPE_TEXT:
	{
	    int mc_depth, mc_type;

	    CDLL_top (&mc_stack, &mc_depth, &mc_type);
	    if (depth != mc_depth + 1 && depth != mc_depth + 2) {
		xmlTextWriterWriteString (writer, val);
	    } 
	    break;
	}

	case XML_READER_TYPE_DOCUMENT_TYPE:
#if 0
	    xmlTextWriterWriteFormatDTD (writer, name, NULL, "TBCppA.dtd",
					 NULL);
	    xmlTextWriterWriteFormatRaw (writer, "\n");
#else
	    {
		char buf [1024];
		xmlDocPtr doc = xmlTextReaderCurrentDoc (reader);
		xmlDtdPtr dtd = xmlGetIntSubset (doc);
		/* .doctype will be used in 'tstrip.xsl' */
		FILE *fp = fopen (".doctype", "wb+");

		xmlElemDump (fp, doc, (xmlNodePtr)dtd);
		fflush (fp);
		rewind (fp);
		while (fgets (buf, sizeof (buf), fp) != NULL) {
		    xmlTextWriterWriteFormatRaw (writer, "%s", buf);
		}
		fclose (fp);
		xmlTextWriterWriteFormatRaw (writer, "\n");
	    }
#endif
	    break;
	case XML_READER_TYPE_SIGNIFICANT_WHITESPACE:
	    xmlTextWriterWriteFormatRaw (writer, "%s", val);
	    break;

	case XML_READER_TYPE_COMMENT:
	    xmlTextWriterWriteComment (writer, val);
	    break;

	case XML_READER_TYPE_ENTITY_REFERENCE:
	    xmlTextWriterWriteFormatRaw (writer, "&%s;", name);
	    break;

	case XML_READER_TYPE_ENTITY:
	case XML_READER_TYPE_NONE:
	case XML_READER_TYPE_ATTRIBUTE:
	case XML_READER_TYPE_XML_DECLARATION:
	case XML_READER_TYPE_CDATA:
	case XML_READER_TYPE_PROCESSING_INSTRUCTION:
	case XML_READER_TYPE_DOCUMENT:
	case XML_READER_TYPE_DOCUMENT_FRAGMENT:
	case XML_READER_TYPE_NOTATION:
	case XML_READER_TYPE_WHITESPACE:
	case XML_READER_TYPE_END_ENTITY:
	    fprintf (stderr, "tfix: unsupported type: %d\n", type);
	    assert (0);
	    break;
	default:
	    fprintf (stderr, "tfix: unknown type: %d\n", type);
	    assert (0);
	    break;
	}

	xmlFree (name);
	xmlFree (val);
    }

    xmlTextWriterWriteFormatRaw (writer, "\n");
    xmlFreeTextReader (reader);
    xmlFreeTextWriter (writer);
}

/* Don't forget to do 'xmlFree" for a pointer of the type "xmlChar". */
/* ==================================================================== */
int
main (int argc, char *argv [])
{
    if (argc < 2) {
	fprintf (stderr, "Usage: %s out.xml < in.xml\n", argv [0]);
	exit (1);
    } else if (argc >= 3) {
	int fd = open (argv [2], O_RDONLY);
	dup2 (fd, STDIN_FILENO);
	close (fd);
    }

    CDLL_init (&mc_stack);
    tfix (argv [1]);
    return 0;
}

